@@ -46,7 +46,7 @@ pub struct ApiError {
 
 impl Display for ApiError {
     fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
-        write!(f, "API error: (code: {}): {}", self.code, self.text)
+        write!(f, "(code: {}): {}", self.code, self.text)
     }
 }
 
@@ -57,19 +57,19 @@ pub enum Error {
     /* Request related errors */
     /// A HTTP error like a 4XX or 5XX status code
     #[cfg(feature = "from-reqwest")]
-    #[error("HTTP error")]
+    #[error("HTTP error: {0}")]
     HttpError(#[from] reqwest::Error),
     /// Invalid header value, likely if the provided OAuth2 token
     /// or User-agent are invalid
     #[cfg(feature = "from-reqwest")]
-    #[error("Invalid header value")]
+    #[error("Invalid header value: {0}")]
     InvalidHeaderValue(#[from] reqwest::header::InvalidHeaderValue),
     /// Error when decoding the JSON response from the API
-    #[error("JSON error")]
+    #[error("JSON error: {0}")]
     JsonError(#[from] serde_json::Error),
     /// Error if unable to get request concurrency lock
     #[cfg(feature = "from-tokio")]
-    #[error("Unable to get request lock")]
+    #[error("Unable to get request lock: {0}")]
     LockError(#[from] tokio::sync::AcquireError),
     /// etag header is invalid/missing
     #[error("The etag for this request is missing or invalid")]
@@ -96,6 +96,12 @@ pub enum Error {
     NotLoggedInAsBot,
     #[error("Missing permission: {0}")]
     PermissionsError(String),
+    #[error("Globally blocked: {0}")]
+    GloballyBlocked(String),
+    #[error("Globally range blocked: {0}")]
+    GloballyRangeBlocked(String),
+    #[error("Globally XFF blocked: {0}")]
+    GloballyXFFBlocked(String),
 
     /* Login-related issues */
     #[error("You've made too many recent login attempts.")]
@@ -119,12 +125,23 @@ pub enum Error {
     /// Tripped the spam filter (aka SpamBlacklist)
     #[error("{info}")]
     SpamFilter { info: String, matches: Vec<String> },
+    /// Some save failure happened, but we don't know what it is
+    #[error("Unknown save failure: {0}")]
+    UnknownSaveFailure(Value),
 
-    /* Catchall/generic issues */
+    /* MediaWiki-side issues */
     #[error("{info}, retry-after: {retry_after}")]
     MaxlagError { info: String, retry_after: u32 },
+    /// When MediaWiki is in readonly mode
+    #[error("MediaWiki is readonly: {0}")]
+    Readonly(String),
+    /// An internal MediaWiki exception
+    #[error("Internal error: {0}")]
+    InternalError(ApiError),
+
+    /* Catchall/generic issues */
     /// Any arbitrary error returned by the MediaWiki API
-    #[error("{0}")]
+    #[error("API error: {0}")]
     ApiError(ApiError),
     /// An error where we don't know what to do nor have
     /// information to report back
@@ -142,6 +159,22 @@ impl Error {
                 | Error::SpamFilter { .. }
         )
     }
+
+    /// Whether the issue is related to a block
+    pub fn is_block_error(&self) -> bool {
+        matches!(
+            self,
+            Error::GloballyBlocked(_)
+                | Error::GloballyRangeBlocked(_)
+                | Error::GloballyXFFBlocked(_)
+        )
+    }
+
+    /// Whether the request should be retried, after some suitable backoff
+    /// and likely with some retry limit
+    pub fn should_retry(&self) -> bool {
+        matches!(self, Error::MaxlagError { .. } | Error::Readonly(_))
+    }
 }
 
 #[derive(Deserialize)]
@@ -155,6 +188,18 @@ impl From<ApiError> for Error {
             "assertuserfailed" => Self::NotLoggedIn,
             "assertbotfailed" => Self::NotLoggedInAsBot,
             "badtoken" => Self::BadToken,
+            "globalblocking-ipblocked"
+            | "wikimedia-globalblocking-ipblocked" => {
+                Self::GloballyBlocked(apierr.text)
+            }
+            "globalblocking-ipblocked-range"
+            | "wikimedia-globalblocking-ipblocked-range" => {
+                Self::GloballyRangeBlocked(apierr.text)
+            }
+            "globalblocking-ipblocked-xff"
+            | "wikimedia-globalblocking-ipblocked-xff" => {
+                Self::GloballyXFFBlocked(apierr.text)
+            }
             "login-throttled" => Self::LoginThrottled,
             "maxlag" => Self::MaxlagError {
                 info: apierr.text,
@@ -163,6 +208,7 @@ impl From<ApiError> for Error {
                 retry_after: 0,
             },
             "protectedpage" => Self::ProtectedPage,
+            "readonly" => Self::Readonly(apierr.text),
             "spamblacklist" => {
                 let matches = if let Some(data) = apierr.data {
                     match serde_json::from_value::<SpamFilterData>(
@@ -181,7 +227,13 @@ impl From<ApiError> for Error {
                 }
             }
             "wrongpassword" => Self::WrongPassword,
-            _ => Self::ApiError(apierr),
+            code => {
+                if code.starts_with("internal_api_error_") {
+                    Self::InternalError(apierr)
+                } else {
+                    Self::ApiError(apierr)
+                }
+            }
         }
     }
 }
@@ -218,10 +270,7 @@ mod tests {
             text: "Some description".to_string(),
             data: None,
         };
-        assert_eq!(
-            &apierr.to_string(),
-            "API error: (code: errorcode): Some description"
-        );
+        assert_eq!(&apierr.to_string(), "(code: errorcode): Some description");
     }
 
     #[test]
@@ -235,7 +284,7 @@ mod tests {
                         "example.org"
                     ]
                 }
-            }))
+            })),
         };
         let err = Error::from(apierr);
         if let Error::SpamFilter { matches, .. } = err {
